@@ -163,9 +163,10 @@ again:
 	}
 
 	if (result < 0) {
-		g_set_error (err, G_IO_CHANNEL_ERROR,
-			     G_IO_CHANNEL_ERROR_FAILED,
-			     "Unable to handshake");
+		g_set_error (err, SOUP_SSL_ERROR,
+			     SOUP_SSL_ERROR_HANDSHAKE_FAILED,
+			     "SSL handshake failed: %s",
+			     gnutls_strerror (result));
 		return G_IO_STATUS_ERROR;
 	}
 
@@ -476,7 +477,7 @@ soup_ssl_wrap_iochannel (GIOChannel *sock, gboolean non_blocking,
 		goto THROW_CREATE_ERROR;
 
 	/* See http://bugzilla.gnome.org/show_bug.cgi?id=581342 */
-	if (gnutls_priority_set_direct (session, "NORMAL:!VERS-TLS1.1:!VERS-TLS1.0", NULL) != 0)
+	if (gnutls_priority_set_direct (session, "NORMAL:!VERS-TLS1.2:!VERS-TLS1.1:!VERS-TLS1.0", NULL) != 0)
 		goto THROW_CREATE_ERROR;
 
 	if (gnutls_credentials_set (session, GNUTLS_CRD_CERTIFICATE,
@@ -485,6 +486,11 @@ soup_ssl_wrap_iochannel (GIOChannel *sock, gboolean non_blocking,
 
 	if (type == SOUP_SSL_TYPE_SERVER)
 		gnutls_dh_set_prime_bits (session, DH_BITS);
+	else {
+		// gnutls defaults to requiring at least 768-bit keys,
+		// but there are some lame servers out there...
+		gnutls_dh_set_prime_bits (session, 256);
+	}
 
 	chan = g_slice_new0 (SoupGNUTLSChannel);
 	chan->real_sock = sock;
