@@ -117,6 +117,8 @@ static void auth_manager_authenticate (SoupAuthManager *manager,
 #define SOUP_SESSION_MAX_CONNS_DEFAULT 10
 #define SOUP_SESSION_MAX_CONNS_PER_HOST_DEFAULT 2
 
+#define SOUP_SESSION_MAX_REDIRECTION_COUNT 20
+
 #define SOUP_SESSION_USER_AGENT_BASE "libsoup/" PACKAGE_VERSION
 
 G_DEFINE_ABSTRACT_TYPE (SoupSession, soup_session, G_TYPE_OBJECT)
@@ -999,7 +1001,8 @@ get_property (GObject *object, guint prop_id,
  * context, so you will need to ref it yourself if you want it to
  * outlive its session.
  *
- * Return value: @session's #GMainContext, which may be %NULL
+ * Return value: (transfer none): @session's #GMainContext, which may
+ * be %NULL
  **/
 GMainContext *
 soup_session_get_async_context (SoupSession *session)
@@ -1092,7 +1095,8 @@ auth_manager_authenticate (SoupAuthManager *manager, SoupMessage *msg,
 static void
 redirect_handler (SoupMessage *msg, gpointer user_data)
 {
-	SoupSession *session = user_data;
+	SoupMessageQueueItem *item = user_data;
+	SoupSession *session = item->session;
 	const char *new_loc;
 	SoupURI *new_uri;
 
@@ -1100,6 +1104,12 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 						"Location");
 	g_return_if_fail (new_loc != NULL);
 
+	if (item->redirection_count >= SOUP_SESSION_MAX_REDIRECTION_COUNT) {
+		soup_session_cancel_message (session, msg, SOUP_STATUS_TOO_MANY_REDIRECTS);
+		return;
+	}
+	item->redirection_count++;
+
 	if (msg->status_code == SOUP_STATUS_SEE_OTHER ||
 	    (msg->status_code == SOUP_STATUS_FOUND &&
 	     !SOUP_METHOD_IS_SAFE (msg->method)) ||
@@ -1164,18 +1174,10 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 void
 soup_session_send_queue_item (SoupSession *session,
 			      SoupMessageQueueItem *item,
-			      SoupConnection *conn)
+			      SoupMessageCompletionFn completion_cb)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 
-	if (item->conn) {
-		if (item->conn != conn) {
-			g_object_unref (item->conn);
-			item->conn = g_object_ref (conn);
-		}
-	} else
-		item->conn = g_object_ref (conn);
-
 	if (priv->user_agent) {
 		soup_message_headers_replace (item->msg->request_headers,
 					      "User-Agent", priv->user_agent);
@@ -1190,8 +1192,8 @@ soup_session_send_queue_item (SoupSession *session,
 	}
 
 	g_signal_emit (session, signals[REQUEST_STARTED], 0,
-		       item->msg, soup_connection_get_socket (conn));
-	soup_connection_send_request (conn, item->msg);
+		       item->msg, soup_connection_get_socket (item->conn));
+	soup_connection_send_request (item->conn, item, completion_cb, item);
 }
 
 gboolean
@@ -1250,92 +1252,12 @@ connection_disconnected (SoupConnection *conn, gpointer user_data)
 	g_object_unref (conn);
 }
 
-void
-soup_session_connection_failed (SoupSession *session,
-				SoupConnection *conn, guint status)
-{
-	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
-	SoupSessionHost *host;
-	SoupMessageQueueItem *item;
-	SoupMessage *msg;
-
-	g_mutex_lock (priv->host_lock);
-	host = g_hash_table_lookup (priv->conns, conn);
-	g_mutex_unlock (priv->host_lock);
-
-	if (host)
-		connection_disconnected (conn, session);
-	else if (conn)
-		host = g_object_get_data (G_OBJECT (conn), "SoupSessionHost");
-	if (!host)
-		return;
-
-	if (status == SOUP_STATUS_TRY_AGAIN)
-		return;
-
-	/* Cancel any other messages waiting for a connection to it,
-	 * since they're out of luck.
-	 */
-	g_object_ref (session);
-	for (item = soup_message_queue_first (priv->queue); item; item = soup_message_queue_next (priv->queue, item)) {
-		msg = item->msg;
-		if (SOUP_MESSAGE_IS_STARTING (msg) &&
-		    get_host_for_message (session, msg) == host)
-			soup_session_cancel_message (session, msg, status);
-	}
-	g_object_unref (session);
-}
-
-static void
-tunnel_connected (SoupMessage *msg, gpointer user_data)
-{
-	SoupSession *session = user_data;
-
-	if (SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
-		SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
-		SoupMessageQueueItem *item =
-			soup_message_queue_lookup (priv->queue, msg);
-
-		/* Clear the connection's proxy_uri, since it is now
-		 * (effectively) directly connected.
-		 */
-		g_object_set (item->conn,
-			      SOUP_CONNECTION_PROXY_URI, NULL,
-			      NULL);
-		soup_message_queue_item_unref (item);
-	}
-}
-
-static void
-tunnel_connect_restarted (SoupMessage *msg, gpointer session)
-{
-	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
-	SoupMessageQueueItem *item;
-
-	if (msg->status_code != SOUP_STATUS_PROXY_AUTHENTICATION_REQUIRED)
-		return;
-
-	item = soup_message_queue_lookup (priv->queue, msg);
-	if (!item)
-		return;
-	if (soup_connection_get_state (item->conn) == SOUP_CONNECTION_DISCONNECTED) {
-		/* We got a 407, and the session provided auth and
-		 * restarted the message, but the proxy closed the
-		 * connection, so we need to create a new one. The
-		 * easiest way to do this is to just give up on the
-		 * current msg and conn, and re-run the queue.
-		 */
-		soup_session_cancel_message (session, msg,
-					     SOUP_STATUS_TRY_AGAIN);
-	}
-	soup_message_queue_item_unref (item);
-}
-
 SoupMessageQueueItem *
-soup_session_make_connect_message (SoupSession *session,
-				   SoupAddress *server_addr)
+soup_session_make_connect_message (SoupSession    *session,
+				   SoupConnection *conn)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupAddress *server_addr = soup_connection_get_tunnel_addr (conn);
 	SoupURI *uri;
 	SoupMessage *msg;
 	SoupMessageQueueItem *item;
@@ -1351,56 +1273,19 @@ soup_session_make_connect_message (SoupSession *session,
 
 	/* Call the base implementation of soup_session_queue_message
 	 * directly, to add msg to the SoupMessageQueue and cause all
-	 * the right signals to be emitted. We can't use
-	 * queue_message's callback arg in this case because that's
-	 * actually implemented by the subclasses.
+	 * the right signals to be emitted.
 	 */
-	g_signal_connect (msg, "finished",
-			  G_CALLBACK (tunnel_connected), session);
-	g_signal_connect (msg, "restarted",
-			  G_CALLBACK (tunnel_connect_restarted), session);
 	queue_message (session, msg, NULL, NULL);
 	item = soup_message_queue_lookup (priv->queue, msg);
+	item->conn = g_object_ref (conn);
 	g_object_unref (msg);
+
+	item->conn = g_object_ref (conn);
+	g_signal_emit (session, signals[TUNNELING], 0, conn);
 	return item;
 }
 
-/**
- * soup_session_get_connection:
- * @session: a #SoupSession
- * @item: a #SoupMessageQueueItem
- * @try_pruning: on return, whether or not to try pruning a connection
- * 
- * Tries to find or create a connection for @item; this is an internal
- * method for #SoupSession subclasses.
- *
- * If there is an idle connection to the relevant host available, then
- * that connection will be returned. The connection will be set to
- * %SOUP_CONNECTION_IN_USE, so the caller must call
- * soup_connection_set_state() to set it to %SOUP_CONNECTION_IDLE if
- * it ends up not using the connection right away.
- *
- * If there is no idle connection available, but it is possible to
- * create a new connection, then one will be created and returned
- * (with state %SOUP_CONNECTION_NEW). The caller MUST then call
- * soup_connection_connect_sync() or soup_connection_connect_async()
- * to connect it. If the connection attempt fails, the caller must
- * call soup_session_connection_failed() to tell the session to free
- * the connection.
- *
- * If no connection is available and a new connection cannot be made,
- * soup_session_get_connection() will return %NULL. If @session has
- * the maximum number of open connections open, but does not have the
- * maximum number of per-host connections open to the relevant host,
- * then *@try_pruning will be set to %TRUE. In this case, the caller
- * can call soup_session_try_prune_connection() to close an idle
- * connection, and then try soup_session_get_connection() again. (If
- * calling soup_session_try_prune_connection() wouldn't help, then
- * *@try_pruning is left untouched; it is NOT set to %FALSE.)
- *
- * Return value: a #SoupConnection, or %NULL
- **/
-SoupConnection *
+gboolean
 soup_session_get_connection (SoupSession *session,
 			     SoupMessageQueueItem *item,
 			     gboolean *try_pruning)
@@ -1414,6 +1299,11 @@ soup_session_get_connection (SoupSession *session,
 	int num_pending = 0;
 	SoupURI *uri;
 
+	if (item->conn) {
+		g_return_val_if_fail (soup_connection_get_state (item->conn) != SOUP_CONNECTION_DISCONNECTED, FALSE);
+		return TRUE;
+	}
+
 	g_mutex_lock (priv->host_lock);
 
 	host = get_host_for_message (session, item->msg);
@@ -1421,7 +1311,8 @@ soup_session_get_connection (SoupSession *session,
 		if (soup_connection_get_state (conns->data) == SOUP_CONNECTION_IDLE) {
 			soup_connection_set_state (conns->data, SOUP_CONNECTION_IN_USE);
 			g_mutex_unlock (priv->host_lock);
-			return conns->data;
+			item->conn = g_object_ref (conns->data);
+			return TRUE;
 		} else if (soup_connection_get_state (conns->data) == SOUP_CONNECTION_CONNECTING)
 			num_pending++;
 	}
@@ -1431,18 +1322,18 @@ soup_session_get_connection (SoupSession *session,
 	 */
 	if (num_pending > host->num_messages / 2) {
 		g_mutex_unlock (priv->host_lock);
-		return NULL;
+		return FALSE;
 	}
 
 	if (host->num_conns >= priv->max_conns_per_host) {
 		g_mutex_unlock (priv->host_lock);
-		return NULL;
+		return FALSE;
 	}
 
 	if (priv->num_conns >= priv->max_conns) {
 		*try_pruning = TRUE;
 		g_mutex_unlock (priv->host_lock);
-		return NULL;
+		return FALSE;
 	}
 
 	if (item->proxy_addr) {
@@ -1474,7 +1365,6 @@ soup_session_get_connection (SoupSession *session,
 		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
 		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
 		NULL);
-	g_object_set_data (G_OBJECT (conn), "SoupSessionHost", host);
 	g_signal_connect (conn, "disconnected",
 			  G_CALLBACK (connection_disconnected),
 			  session);
@@ -1488,7 +1378,8 @@ soup_session_get_connection (SoupSession *session,
 	host->connections = g_slist_prepend (host->connections, conn);
 
 	g_mutex_unlock (priv->host_lock);
-	return conn;
+	item->conn = g_object_ref (conn);
+	return TRUE;
 }
 
 SoupMessageQueue *
@@ -1499,11 +1390,10 @@ soup_session_get_queue (SoupSession *session)
 	return priv->queue;
 }
 
-static void
-message_finished (SoupMessage *msg, gpointer user_data)
+void
+soup_session_unqueue_item (SoupSession          *session,
+			   SoupMessageQueueItem *item)
 {
-	SoupMessageQueueItem *item = user_data;
-	SoupSession *session = item->session;
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupSessionHost *host;
 
@@ -1512,23 +1402,61 @@ message_finished (SoupMessage *msg, gpointer user_data)
 		item->conn = NULL;
 	}
 
-	if (!SOUP_MESSAGE_IS_STARTING (msg)) {
-		soup_message_queue_remove (priv->queue, item);
+	if (item->state != SOUP_MESSAGE_FINISHED) {
+		g_warning ("finished an item with state %d", item->state);
+		return;
+	}
 
-		g_mutex_lock (priv->host_lock);
-		host = get_host_for_message (session, item->msg);
-		host->num_messages--;
-		g_mutex_unlock (priv->host_lock);
+	soup_message_queue_remove (priv->queue, item);
 
-		g_signal_handlers_disconnect_by_func (msg, message_finished, item);
-		/* g_signal_handlers_disconnect_by_func doesn't work if you
-		 * have a metamarshal, meaning it doesn't work with
-		 * soup_message_add_header_handler()
-		 */
-		g_signal_handlers_disconnect_matched (msg, G_SIGNAL_MATCH_DATA,
-						      0, 0, NULL, NULL, session);
-		g_signal_emit (session, signals[REQUEST_UNQUEUED], 0, msg);
-		soup_message_queue_item_unref (item);
+	g_mutex_lock (priv->host_lock);
+	host = get_host_for_message (session, item->msg);
+	host->num_messages--;
+	g_mutex_unlock (priv->host_lock);
+
+	/* g_signal_handlers_disconnect_by_func doesn't work if you
+	 * have a metamarshal, meaning it doesn't work with
+	 * soup_message_add_header_handler()
+	 */
+	g_signal_handlers_disconnect_matched (item->msg, G_SIGNAL_MATCH_DATA,
+					      0, 0, NULL, NULL, item);
+	g_signal_emit (session, signals[REQUEST_UNQUEUED], 0, item->msg);
+	soup_message_queue_item_unref (item);
+}
+
+void
+soup_session_set_item_status (SoupSession          *session,
+			      SoupMessageQueueItem *item,
+			      guint                 status_code)
+{
+	SoupURI *uri;
+	char *msg;
+
+	switch (status_code) {
+	case SOUP_STATUS_CANT_RESOLVE:
+	case SOUP_STATUS_CANT_CONNECT:
+		uri = soup_message_get_uri (item->msg);
+		msg = g_strdup_printf ("%s (%s)",
+				       soup_status_get_phrase (status_code),
+				       uri->host);
+		soup_message_set_status_full (item->msg, status_code, msg);
+		g_free (msg);
+		break;
+
+	case SOUP_STATUS_CANT_RESOLVE_PROXY:
+	case SOUP_STATUS_CANT_CONNECT_PROXY:
+		if (item->proxy_uri && item->proxy_uri->host) {
+			msg = g_strdup_printf ("%s (%s)",
+					       soup_status_get_phrase (status_code),
+					       item->proxy_uri->host);
+			soup_message_set_status_full (item->msg, status_code, msg);
+			g_free (msg);
+			break;
+		}
+		/* else fall through */
+
+	default:
+		soup_message_set_status (item->msg, status_code);
 	}
 }
 
@@ -1547,15 +1475,10 @@ queue_message (SoupSession *session, SoupMessage *msg,
 	host->num_messages++;
 	g_mutex_unlock (priv->host_lock);
 
-	soup_message_set_io_status (msg, SOUP_MESSAGE_IO_STATUS_QUEUED);
-
-	g_signal_connect_after (msg, "finished",
-				G_CALLBACK (message_finished), item);
-
 	if (!(soup_message_get_flags (msg) & SOUP_MESSAGE_NO_REDIRECT)) {
 		soup_message_add_header_handler (
 			msg, "got_body", "Location",
-			G_CALLBACK (redirect_handler), session);
+			G_CALLBACK (redirect_handler), item);
 	}
 
 	g_signal_emit (session, signals[REQUEST_QUEUED], 0, msg);
@@ -1602,7 +1525,13 @@ soup_session_queue_message (SoupSession *session, SoupMessage *msg,
 static void
 requeue_message (SoupSession *session, SoupMessage *msg)
 {
-	soup_message_set_io_status (msg, SOUP_MESSAGE_IO_STATUS_QUEUED);
+	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupMessageQueueItem *item;
+
+	item = soup_message_queue_lookup (priv->queue, msg);
+	g_return_if_fail (item != NULL);
+	item->state = SOUP_MESSAGE_RESTARTING;
+	soup_message_queue_item_unref (item);
 }
 
 /**
@@ -1694,15 +1623,18 @@ cancel_message (SoupSession *session, SoupMessage *msg, guint status_code)
 	SoupMessageQueueItem *item;
 
 	item = soup_message_queue_lookup (priv->queue, msg);
-	if (item) {
-		if (item->cancellable)
-			g_cancellable_cancel (item->cancellable);
-		soup_message_queue_item_unref (item);
-	}
+	g_return_if_fail (item != NULL);
+
+	if (item->cancellable)
+		g_cancellable_cancel (item->cancellable);
 
-	soup_message_io_stop (msg);
 	soup_message_set_status (msg, status_code);
-	soup_message_finished (msg);
+	if (soup_message_io_in_progress (msg))
+		soup_message_io_finished (msg);
+	else
+		item->state = SOUP_MESSAGE_FINISHING;
+
+	soup_message_queue_item_unref (item);
 }
 
 /**
@@ -1735,14 +1667,24 @@ void
 soup_session_cancel_message (SoupSession *session, SoupMessage *msg,
 			     guint status_code)
 {
+	SoupSessionPrivate *priv;
+	SoupMessageQueueItem *item;
+
 	g_return_if_fail (SOUP_IS_SESSION (session));
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
+	priv = SOUP_SESSION_GET_PRIVATE (session);
+	item = soup_message_queue_lookup (priv->queue, msg);
 	/* If the message is already ending, don't do anything */
-	if (soup_message_get_io_status (msg) == SOUP_MESSAGE_IO_STATUS_FINISHED)
+	if (!item)
 		return;
+	if (item->state == SOUP_MESSAGE_FINISHED) {
+		soup_message_queue_item_unref (item);
+		return;
+	}
 
 	SOUP_SESSION_GET_CLASS (session)->cancel_message (session, msg, status_code);
+	soup_message_queue_item_unref (item);
 }
 
 static void
@@ -1946,8 +1888,8 @@ restart:
  * you want to see all features, you can pass %G_TYPE_SESSION_FEATURE
  * for @feature_type.)
  *
- * Return value: a list of features. You must free the list, but not
- * its contents
+ * Return value: (transfer container): a list of features. You must
+ * free the list, but not its contents
  *
  * Since: 2.26
  **/
@@ -1976,8 +1918,8 @@ soup_session_get_features (SoupSession *session, GType feature_type)
  * features where there may be more than one feature of a given type,
  * use soup_session_get_features().
  *
- * Return value: a #SoupSessionFeature, or %NULL. The feature is owned
- * by @session.
+ * Return value: (transfer none): a #SoupSessionFeature, or %NULL. The
+ * feature is owned by @session.
  *
  * Since: 2.26
  **/
@@ -2023,8 +1965,8 @@ soup_session_get_feature (SoupSession *session, GType feature_type)
  * disabled on @msg, and the second is not, then this will return
  * %NULL, not the second feature.
  *
- * Return value: a #SoupSessionFeature, or %NULL. The feature is owned
- * by @session.
+ * Return value: (transfer none): a #SoupSessionFeature, or %NULL. The
+ * feature is owned by @session.
  *
  * Since: 2.28
  **/
