@@ -35,7 +35,18 @@
  * soup_socket_get_remote_address()) may be useful to applications.
  **/
 
-G_DEFINE_TYPE (SoupSocket, soup_socket, G_TYPE_OBJECT)
+static void
+_soup_networking_init (void)
+{
+#ifdef G_OS_WIN32
+  WSADATA wsadata;
+  if (WSAStartup (MAKEWORD (2, 0), &wsadata) != 0)
+    g_error ("Windows Sockets could not be initialized");
+#endif
+}
+
+G_DEFINE_TYPE_WITH_CODE (SoupSocket, soup_socket, G_TYPE_OBJECT,
+			 _soup_networking_init ();)
 
 enum {
 	READABLE,
@@ -55,8 +66,11 @@ enum {
 	PROP_NON_BLOCKING,
 	PROP_IS_SERVER,
 	PROP_SSL_CREDENTIALS,
+	PROP_SSL_STRICT,
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
+	PROP_TRUSTED_CERTIFICATE,
+	PROP_CLEAN_DISPOSE,
 
 	LAST_PROP
 };
@@ -69,6 +83,9 @@ typedef struct {
 	guint non_blocking:1;
 	guint is_server:1;
 	guint timed_out:1;
+	guint ssl_strict:1;
+	guint trusted_certificate:1;
+	guint clean_dispose:1;
 	gpointer ssl_creds;
 
 	GMainContext   *async_context;
@@ -80,15 +97,11 @@ typedef struct {
 
 	GMutex *iolock, *addrlock;
 	guint timeout;
+
+	GCancellable *connect_cancel;
 } SoupSocketPrivate;
 #define SOUP_SOCKET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SOCKET, SoupSocketPrivate))
 
-#ifdef HAVE_IPV6
-#define soup_sockaddr_max sockaddr_in6
-#else
-#define soup_sockaddr_max sockaddr_in
-#endif
-
 static void set_property (GObject *object, guint prop_id,
 			  const GValue *value, GParamSpec *pspec);
 static void get_property (GObject *object, guint prop_id,
@@ -120,8 +133,10 @@ soup_socket_init (SoupSocket *sock)
 static void
 disconnect_internal (SoupSocketPrivate *priv)
 {
-	g_io_channel_unref (priv->iochannel);
-	priv->iochannel = NULL;
+	if (priv->iochannel) {
+		g_io_channel_unref (priv->iochannel);
+		priv->iochannel = NULL;
+	}
 	priv->sockfd = -1;
 
 	if (priv->read_src) {
@@ -147,18 +162,32 @@ finalize (GObject *object)
 {
 	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (object);
 
-	if (priv->iochannel)
+	if (priv->connect_cancel) {
+		if (priv->clean_dispose)
+			g_warning ("Disposing socket %p during connect", object);
+		g_object_unref (priv->connect_cancel);
+	}
+	if (priv->iochannel) {
+		if (priv->clean_dispose)
+			g_warning ("Disposing socket %p while still connected", object);
 		disconnect_internal (priv);
+	}
 
 	if (priv->local_addr)
 		g_object_unref (priv->local_addr);
 	if (priv->remote_addr)
 		g_object_unref (priv->remote_addr);
 
-	if (priv->watch_src)
+	if (priv->watch_src) {
+		if (priv->clean_dispose && !priv->is_server)
+			g_warning ("Disposing socket %p during async op", object);
 		g_source_destroy (priv->watch_src);
-	if (priv->connect_timeout)
+	}
+	if (priv->connect_timeout) {
+		if (priv->clean_dispose)
+			g_warning ("Disposing socket %p during connect (2)", object);
 		g_source_destroy (priv->connect_timeout);
+	}
 	if (priv->async_context)
 		g_main_context_unref (priv->async_context);
 
@@ -348,6 +377,33 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 				      "SSL credentials",
 				      "SSL credential information, passed from the session to the SSL implementation",
 				      G_PARAM_READWRITE));
+	/**
+	 * SOUP_SOCKET_SSL_STRICT:
+	 *
+	 * Alias for the #SoupSocket:ignore-ssl-cert-errors property.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_SSL_STRICT,
+		g_param_spec_boolean (SOUP_SOCKET_SSL_STRICT,
+				      "Strictly validate SSL certificates",
+				      "Whether certificate errors should be considered a connection error",
+				      TRUE,
+				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SOCKET_TRUSTED_CERTIFICATE:
+	 *
+	 * Alias for the #SoupSocket:trusted-certificate
+	 * property. Notice that this property's value is only useful
+	 * if the socket is for an SSL connection, and only reliable
+	 * after some data has been transferred to or from it.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_TRUSTED_CERTIFICATE,
+		g_param_spec_boolean (SOUP_SOCKET_TRUSTED_CERTIFICATE,
+				     "Trusted Certificate",
+				     "Whether the server certificate is trusted, if this is an SSL socket",
+				     FALSE,
+				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	/**
 	 * SOUP_SOCKET_ASYNC_CONTEXT:
 	 *
@@ -375,10 +431,13 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 				   0, G_MAXUINT, 0,
 				   G_PARAM_READWRITE));
 
-#ifdef G_OS_WIN32
-	/* Make sure WSAStartup() gets called. */
-	soup_address_get_type ();
-#endif
+	g_object_class_install_property (
+		object_class, PROP_CLEAN_DISPOSE,
+		g_param_spec_boolean ("clean-dispose",
+				      "Clean dispose",
+				      "Warn on unclean dispose",
+				      FALSE,
+				      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
 }
 
 
@@ -492,6 +551,12 @@ set_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDENTIALS:
 		priv->ssl_creds = g_value_get_pointer (value);
 		break;
+	case PROP_SSL_STRICT:
+		priv->ssl_strict = g_value_get_boolean (value);
+		break;
+	case PROP_TRUSTED_CERTIFICATE:
+		priv->trusted_certificate = g_value_get_boolean (value);
+		break;
 	case PROP_ASYNC_CONTEXT:
 		priv->async_context = g_value_get_pointer (value);
 		if (priv->async_context)
@@ -500,6 +565,9 @@ set_property (GObject *object, guint prop_id,
 	case PROP_TIMEOUT:
 		priv->timeout = g_value_get_uint (value);
 		break;
+	case PROP_CLEAN_DISPOSE:
+		priv->clean_dispose = g_value_get_boolean (value);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -528,6 +596,12 @@ get_property (GObject *object, guint prop_id,
 	case PROP_SSL_CREDENTIALS:
 		g_value_set_pointer (value, priv->ssl_creds);
 		break;
+	case PROP_SSL_STRICT:
+		g_value_set_boolean (value, priv->ssl_strict);
+		break;
+	case PROP_TRUSTED_CERTIFICATE:
+		g_value_set_boolean (value, priv->trusted_certificate);
+		break;
 	case PROP_ASYNC_CONTEXT:
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
@@ -572,6 +646,16 @@ typedef struct {
 	gpointer user_data;
 } SoupSocketAsyncConnectData;
 
+static void
+free_sacd (SoupSocketAsyncConnectData *sacd)
+{
+	if (sacd->cancel_id)
+		g_signal_handler_disconnect (sacd->cancellable, sacd->cancel_id);
+	g_object_unref (sacd->cancellable);
+	g_object_unref (sacd->sock);
+	g_slice_free (SoupSocketAsyncConnectData, sacd);
+}
+
 static gboolean
 idle_connect_result (gpointer user_data)
 {
@@ -579,9 +663,8 @@ idle_connect_result (gpointer user_data)
 	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sacd->sock);
 	guint status;
 
+	priv->connect_cancel = NULL;
 	priv->watch_src = NULL;
-	if (sacd->cancel_id)
-		g_signal_handler_disconnect (sacd->cancellable, sacd->cancel_id);
 
 	if (priv->sockfd == -1) {
 		if (g_cancellable_is_cancelled (sacd->cancellable))
@@ -591,8 +674,14 @@ idle_connect_result (gpointer user_data)
 	} else
 		status = SOUP_STATUS_OK;
 
+	/* Have to do this before calling the callback... */
+	if (sacd->cancel_id) {
+		g_signal_handler_disconnect (sacd->cancellable, sacd->cancel_id);
+		sacd->cancel_id = 0;
+	}
+
 	sacd->callback (sacd->sock, status, sacd->user_data);
-	g_slice_free (SoupSocketAsyncConnectData, sacd);
+	free_sacd (sacd);
 	return FALSE;
 }
 
@@ -641,16 +730,17 @@ static void
 got_address (SoupAddress *addr, guint status, gpointer user_data)
 {
 	SoupSocketAsyncConnectData *sacd = user_data;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sacd->sock);
 
-	if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
+	priv->connect_cancel = NULL;
+
+	if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
+		soup_socket_connect_async (sacd->sock, sacd->cancellable,
+					   sacd->callback, sacd->user_data);
+	} else
 		sacd->callback (sacd->sock, status, sacd->user_data);
-		g_slice_free (SoupSocketAsyncConnectData, sacd);
-		return;
-	}
 
-	soup_socket_connect_async (sacd->sock, sacd->cancellable,
-				   sacd->callback, sacd->user_data);
-	g_slice_free (SoupSocketAsyncConnectData, sacd);
+	free_sacd (sacd);
 }
 
 static void
@@ -707,7 +797,7 @@ socket_connect_internal (SoupSocket *sock)
  * soup_socket_connect_async:
  * @sock: a client #SoupSocket (which must not already be connected)
  * @cancellable: a #GCancellable, or %NULL
- * @callback: callback to call after connecting
+ * @callback: (scope async): callback to call after connecting
  * @user_data: data to pass to @callback
  *
  * Begins asynchronously connecting to @sock's remote address. The
@@ -731,15 +821,23 @@ soup_socket_connect_async (SoupSocket *sock, GCancellable *cancellable,
 	g_return_if_fail (priv->remote_addr != NULL);
 
 	sacd = g_slice_new0 (SoupSocketAsyncConnectData);
-	sacd->sock = sock;
-	sacd->cancellable = cancellable;
+	sacd->sock = g_object_ref (sock);
+	sacd->cancellable = cancellable ? g_object_ref (cancellable) : g_cancellable_new ();
 	sacd->callback = callback;
 	sacd->user_data = user_data;
 
+	priv->connect_cancel = sacd->cancellable;
+
+	if (g_cancellable_is_cancelled (priv->connect_cancel)) {
+		priv->watch_src = soup_add_completion (priv->async_context,
+						       idle_connect_result, sacd);
+		return;
+	}
+
 	if (!soup_address_get_sockaddr (priv->remote_addr, NULL)) {
 		soup_address_resolve_async (priv->remote_addr,
 					    priv->async_context,
-					    cancellable,
+					    sacd->cancellable,
 					    got_address, sacd);
 		return;
 	}
@@ -760,12 +858,10 @@ soup_socket_connect_async (SoupSocket *sock, GCancellable *cancellable,
 						  priv->timeout * 1000,
 						  connect_timeout, sacd);
 		}
-		if (cancellable) {
-			sacd->cancel_id =
-				g_signal_connect (cancellable, "cancelled",
-						  G_CALLBACK (async_cancel),
-						  sacd);
-		}
+		sacd->cancel_id =
+			g_signal_connect (sacd->cancellable, "cancelled",
+					  G_CALLBACK (async_cancel),
+					  sacd);
 	} else {
 		priv->watch_src = soup_add_completion (priv->async_context,
 						       idle_connect_result, sacd);
@@ -805,28 +901,35 @@ soup_socket_connect_sync (SoupSocket *sock, GCancellable *cancellable)
 	g_return_val_if_fail (priv->sockfd == -1, SOUP_STATUS_MALFORMED);
 	g_return_val_if_fail (priv->remote_addr != NULL, SOUP_STATUS_MALFORMED);
 
+	if (cancellable)
+		g_object_ref (cancellable);
+	else
+		cancellable = g_cancellable_new ();
+	priv->connect_cancel = cancellable;
+
 	if (!soup_address_get_sockaddr (priv->remote_addr, NULL)) {
 		status = soup_address_resolve_sync (priv->remote_addr,
 						    cancellable);
-		if (!SOUP_STATUS_IS_SUCCESSFUL (status))
+		if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
+			priv->connect_cancel = NULL;
+			g_object_unref (cancellable);
 			return status;
+		}
 	}
 
-	if (cancellable) {
-		cancel_id = g_signal_connect (cancellable, "cancelled",
-					      G_CALLBACK (sync_cancel), sock);
-	}
+	cancel_id = g_signal_connect (cancellable, "cancelled",
+				      G_CALLBACK (sync_cancel), sock);
 
 	status = socket_connect_internal (sock);
+	priv->connect_cancel = NULL;
 
-	if (cancellable) {
-		if (status != SOUP_STATUS_OK &&
-		    g_cancellable_is_cancelled (cancellable)) {
-			status = SOUP_STATUS_CANCELLED;
-			disconnect_internal (priv);
-		}
-		g_signal_handler_disconnect (cancellable, cancel_id);
+	if (status != SOUP_STATUS_OK &&
+	    g_cancellable_is_cancelled (cancellable)) {
+		status = SOUP_STATUS_CANCELLED;
+		disconnect_internal (priv);
 	}
+	g_signal_handler_disconnect (cancellable, cancel_id);
+	g_object_unref (cancellable);
 
 	return status;
 }
@@ -844,7 +947,7 @@ listen_watch (GIOChannel* iochannel, GIOCondition condition, gpointer data)
 {
 	SoupSocket *sock = data, *new;
 	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock), *new_priv;
-	struct soup_sockaddr_max sa;
+	struct sockaddr_storage sa;
 	int sa_len, sockfd;
 
 	if (condition & (G_IO_HUP | G_IO_ERR)) {
@@ -991,6 +1094,12 @@ soup_socket_start_proxy_ssl (SoupSocket *sock, const char *ssl_host,
 	if (!ssl_chan)
 		return FALSE;
 
+	/* This is optimistic, we will set this to false if we get a
+	 * cert error from one of the I/O calls
+	 */
+	if (priv->ssl_creds)
+		priv->trusted_certificate = TRUE;
+
 	priv->iochannel = ssl_chan;
 	g_io_channel_unref (real_chan);
 
@@ -1031,7 +1140,10 @@ soup_socket_disconnect (SoupSocket *sock)
 	g_return_if_fail (SOUP_IS_SOCKET (sock));
 	priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
-	if (g_mutex_trylock (priv->iolock)) {
+	if (priv->connect_cancel) {
+		g_cancellable_cancel (priv->connect_cancel);
+		return;
+	} else if (g_mutex_trylock (priv->iolock)) {
 		if (priv->iochannel)
 			disconnect_internal (priv);
 		else
@@ -1058,6 +1170,11 @@ soup_socket_disconnect (SoupSocket *sock)
 	if (already_disconnected)
 		return;
 
+	/* Keep ref around signals in case the object is unreferenced
+	 * in a handler
+	 */
+	g_object_ref (sock);
+
 	/* Give all readers a chance to notice the connection close */
 	g_signal_emit (sock, signals[READABLE], 0);
 
@@ -1065,6 +1182,8 @@ soup_socket_disconnect (SoupSocket *sock)
 
 	/* Then let everyone know we're disconnected */
 	g_signal_emit (sock, signals[DISCONNECTED], 0);
+
+	g_object_unref (sock);
 }
 
 /**
@@ -1092,7 +1211,7 @@ soup_socket_is_connected (SoupSocket *sock)
  *
  * Returns the #SoupAddress corresponding to the local end of @sock.
  *
- * Return value: the #SoupAddress
+ * Return value: (transfer none): the #SoupAddress
  **/
 SoupAddress *
 soup_socket_get_local_address (SoupSocket *sock)
@@ -1104,7 +1223,7 @@ soup_socket_get_local_address (SoupSocket *sock)
 
 	g_mutex_lock (priv->addrlock);
 	if (!priv->local_addr) {
-		struct soup_sockaddr_max bound_sa;
+		struct sockaddr_storage bound_sa;
 		int sa_len;
 
 		sa_len = sizeof (bound_sa);
@@ -1122,7 +1241,7 @@ soup_socket_get_local_address (SoupSocket *sock)
  *
  * Returns the #SoupAddress corresponding to the remote end of @sock.
  *
- * Return value: the #SoupAddress
+ * Return value: (transfer none): the #SoupAddress
  **/
 SoupAddress *
 soup_socket_get_remote_address (SoupSocket *sock)
@@ -1134,7 +1253,7 @@ soup_socket_get_remote_address (SoupSocket *sock)
 
 	g_mutex_lock (priv->addrlock);
 	if (!priv->remote_addr) {
-		struct soup_sockaddr_max bound_sa;
+		struct sockaddr_storage bound_sa;
 		int sa_len;
 
 		sa_len = sizeof (bound_sa);
@@ -1205,14 +1324,26 @@ read_from_network (SoupSocket *sock, gpointer buffer, gsize len,
 	if (!priv->iochannel)
 		return SOUP_SOCKET_EOF;
 
-	if (priv->timed_out)
+	if (priv->timed_out) {
+		g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT,
+				     "Timed out");
 		return SOUP_SOCKET_ERROR;
+	}
 
+again:
 	status = g_io_channel_read_chars (priv->iochannel,
 					  buffer, len, nread, &my_err);
 	if (my_err) {
-		if (my_err->domain == SOUP_SSL_ERROR &&
-		    my_err->code == SOUP_SSL_ERROR_HANDSHAKE_NEEDS_WRITE)
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_CERTIFICATE) &&
+		    !priv->ssl_strict) {
+			priv->trusted_certificate = FALSE;
+			g_clear_error (&my_err);
+			goto again;
+		}
+
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_HANDSHAKE_NEEDS_WRITE))
 			cond = G_IO_OUT;
 		g_propagate_error (error, my_err);
 	}
@@ -1229,8 +1360,11 @@ read_from_network (SoupSocket *sock, gpointer buffer, gsize len,
 		 * a socket timeout and should be treated as an error
 		 * condition.
 		 */
-		if (!priv->non_blocking)
+		if (!priv->non_blocking) {
+			g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT,
+					     "Timed out");
 			return SOUP_SOCKET_ERROR;
+		}
 
 		if (!priv->read_src) {
 			priv->read_src =
@@ -1495,6 +1629,8 @@ soup_socket_write (SoupSocket *sock, gconstpointer buffer,
 	}
 	if (priv->timed_out) {
 		g_mutex_unlock (priv->iolock);
+		g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT,
+				     "Timed out");
 		return SOUP_SOCKET_ERROR;
 	}
 	if (priv->write_src) {
@@ -1502,11 +1638,20 @@ soup_socket_write (SoupSocket *sock, gconstpointer buffer,
 		return SOUP_SOCKET_WOULD_BLOCK;
 	}
 
+again:
 	status = g_io_channel_write_chars (priv->iochannel,
 					   buffer, len, nwrote, &my_err);
 	if (my_err) {
-		if (my_err->domain == SOUP_SSL_ERROR &&
-		    my_err->code == SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ)
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_CERTIFICATE) &&
+		    !priv->ssl_strict) {
+			priv->trusted_certificate = FALSE;
+			g_clear_error (&my_err);
+			goto again;
+		}
+
+		if (g_error_matches (my_err, SOUP_SSL_ERROR,
+				     SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ))
 			cond = G_IO_IN;
 		g_propagate_error (error, my_err);
 	}
@@ -1516,6 +1661,8 @@ soup_socket_write (SoupSocket *sock, gconstpointer buffer,
 	 */
 	if (!priv->non_blocking && status == G_IO_STATUS_AGAIN) {
 		g_mutex_unlock (priv->iolock);
+		g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT,
+				     "Timed out");
 		return SOUP_SOCKET_ERROR;
 	}
 
